/*--------------------------------------------------------------------------
     "libfunc.cc" - library functions and classes
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Aug.   5, 2002
                                        edited: May    1, 2004
                                        edited: Jan.  26, 2011
                                   last edited: May   21, 2021
--------------------------------------------------------------------------*/

#include "libfunc.h"

// =========================================================================
//        general purpose functions and classes
// =========================================================================

void showRevision() { cout << REVSTR << endl; }

/*===== string formatting and parsing functions ========================*/

string strToUpper(const string &s) {
  int ic;
  string rs = s;
  for(ic = 0; ic < rs.length(); ic++)
    { if (islower(rs[ic])) { rs[ic] = toupper(rs[ic]); } }
  return rs;
}

string strToLower(const string &s) {
  int ic;
  string rs = s;
  for(ic = 0; ic < rs.length(); ic++)
    { if (isupper(rs[ic])) { rs[ic] = tolower(rs[ic]); } }
  return rs;
}

string format(int v) {
  ostringstream os;
  os << v;
  return os.str();
}

string format(int v, int len) {
  ostringstream os;
  os << setw(len) << setfill('0') << v;
  return os.str();
}

string format(string s, int len) {
  ostringstream os;
  os << setw(len) << setfill(' ') << s.c_str();
  return os.str();
}

string formatLeft(string s, int len) {
  ostringstream os;
  os << setw(len) << setiosflags(ios::left) << s.c_str();
  return os.str();
}

string formatRight(string s, int len) {
  ostringstream os;
  os << setw(len) << setiosflags(ios::right) << s.c_str();
  return os.str();
}

string formatE(double v, int len, int prc) {
  ostringstream os;
  string rs;
  int posE;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::scientific) << setw(len) << setprecision(prc)
     << v;
  //--- for compiler that produces e+000 output ---
  rs = os.str(); posE = rs.find('e');
  if ((rs.length() - posE) > 4) {
    if (rs[posE + 2] == '0') {
      rs = rs.substr(0, posE + 2) + rs.substr(posE + 3);
      if (rs.length() < len) { rs = ' ' + rs; }
    }
  }
  return rs;
}

string formatF(double v, int len, int prc) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::fixed) << setw(len) << setprecision(prc)
     << v;
  return os.str();
}

string formatFnoDot(double v, int len) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::fixed)
     << setw(len) << setprecision(0) << v;
  return os.str();
}

string formatFautoLen(double v, int prc) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::fixed) << setprecision(prc) << v;
  return os.str();
}

string formatD(int v, int len) {
  ostringstream os;
  os << setw(len) << setfill(' ') << v;
  return os.str();
}

string formatBool(int flag) {
  if (flag) { return "Y"; } else { return "N"; }
}

int valBoolStr(string s) {
  if ( (s == "Y") || (s == "y") || (s == "YES") || (s == "Yes") ||
       (s == "yes") ) { return true; }
  else { return false; }
}

long valHexStr(string s) {
  static const string hxtbl = "0123456789abcdef";
  char c; int ic, dig;
  long v = 0L;
  for (ic = 0; ic < s.length(); ic++) {
    c = s[ic]; if (isupper(c)) { c = tolower(c); }
    if ((dig = hxtbl.find(c)) == string::npos) { dig = 0; }
    v <<= 4; v += dig;
  }
  return v;
}

double atofx(const string &s) {
  char c;
  int lens = s.length(), ic;
  string sx = "";
  for (ic = 0; ic < lens; ic++) {
    c = s[ic];
    if (c == 'D' || c == 'd') { sx += 'E'; }
    else  { sx += c; }
  }
  return atof(sx.c_str());
}

string strCurDate() {
char sdate[20];
time_t curTime;
  time(&curTime);
  strftime(sdate, (size_t)19, "%b %d, %Y", localtime(&curTime));
  return sdate;
}

int nextCorCket(string &s, int start) {
  int lev = 0, retPos = -1;
  for (int i = start; i < s.length(); i++) {
    if (s[i] == ')') {
      if (lev == 0) { retPos = i; break; }
      else { lev--; }
    } else if (s[i] == '(') {
      lev++;
    }
  }
  return retPos;
}

int numMatchStrList(string s, vector<string> *vs) {
  int nmat = 0;
  for (int i = 0; i < vs->size(); i++) {
    if ((*vs)[i] == s) { nmat++; }
  }
  return nmat;
}

ostream &operator<<(ostream &os, vector<string> &vs) {
  vector<string>::iterator p;
  for (p = vs.begin(); p != vs.end(); p++) { os << *p << endl; }
  return os;
}

//----- split tokens in string s ---------------------------------------

int splitToken(const string &s, const string &spl, vector<string> &vs) {
  char c;
  int pTkn, pLastTkn = -1, lens = s.length(), ic, lss;
  if (!vs.empty()) { vs.clear(); }
  for (ic = 0; ic < lens; ic++) {
    c = s[ic];
    if (spl.find(c) != string::npos) {
      pTkn = ic; lss = pTkn - pLastTkn - 1;
      if (lss > 0) { vs.push_back(s.substr(pLastTkn + 1, lss)); }
      pLastTkn = pTkn;
    }
  }
  if (pLastTkn < (lens - 1)) {
    pTkn = lens; lss = pTkn - pLastTkn - 1;
    vs.push_back(s.substr(pLastTkn + 1, lss));
  }
  return vs.size();
}

//----- string terminated by the specified char ------------------------

string strTermBy(const string &s, char ct) {
  int tpos;
  tpos = s.find(ct);
  return s.substr(0, tpos);
}

//----- remove preceding and subsequent space chars --------------------

string strRemoveSpc(const string &s) {
  int ic, pstart, pend, slen;
  string rs;
  slen = s.length(); pstart = slen; pend = -1;
  for (ic = 0; ic < slen; ic++) {
    if (s[ic] != ' ') { pstart = ic; break; }
  }
  for (ic = slen - 1; ic >= 0; ic--) {
    if (s[ic] != ' ') { pend = ic; break; }
  }
  if (pend < pstart) { rs = ""; }
  else { rs = s.substr(pstart, pend - pstart + 1); }
  return rs;
}

//----- get one line from the file stream ------------------------------

int getOneLine(ifstream &inf, string &s) {
  char c;
  if (s != "") { s = ""; }
  while (!inf.get(c).eof()) {
    if (c == '\n') { break; }
    if (!iscntrl(c)) { s += c; }
  }
  if (inf.eof() && (s == "")) { return ERRE; }
  return NORM;
}

//----- get a line from stream with removal of comment part ------------

int getLineRmCmm(ifstream &inf, string &s, const string &cms) {
  int ic;
  char c;
  string::size_type p;
  do {
    if (getOneLine(inf, s) == ERRE) { return ERRE; }
    for (ic = 0; ic < cms.length(); ic++) {
      c = cms[ic];
      if ((p = s.find(c)) != string::npos)
        { s = s.substr(0, p); }
    }
    s = strRemoveSpc(s);
  } while (s == "");
  return NORM;
}

//----- get a line from stream removing comment part (#!) --------------

int getLineApp(ifstream &inf, string &s) {
  return getLineRmCmm(inf, s, "#!");
}

//----- get net ntkn tokens from the file stream -----------------------

int getTokensFrom(ifstream &inf, int ntkn, vector<string> &vs) {
  int itk;
  string s;
  if (!vs.empty()) { vs.clear(); }
  for (itk = 0; itk < ntkn; itk++) {
    if (getTokenFrom(inf, s) == ERRE) { return ERRE; }
    vs.push_back(s);
  }
  return NORM;
}

/*----- get the next token from the file stream ------------------------
     ignoring comments ('!' to line-end)
----------------------------------------------------------------------*/

int getTokenFrom(ifstream &inf, string &s) {
  char c;
  int inComment = false, inToken = false, doneToken = false;
  static string tokenSep = " ,\t\n\r";
  s = "";
  if (inf.eof()) { return ERRE; }
  while (!inf.get(c).eof()) {
    if (inComment) {
      if (c == '\n') {
        inComment = false;
        if (doneToken) { break; }
      }
    } else {
      if (c == '!') {
        if (inToken) { doneToken = true; }
        inComment = true;
      } else {
        if ((tokenSep.find(c) != string::npos) || iscntrl(c)) {
          if (inToken) { doneToken = true; break; }
        } else {
          if (!inToken) { inToken = true; }
          s += c;
        }
      }
    }
  }
  if (!doneToken) { return ERRE; }
  return NORM;
}

/*----- get all characters from a file ---------------------------------
     ignoring comments ('!' to line-end), space, and control codes
----------------------------------------------------------------------*/

int getCharsFile(string flname, string &s) {
  char c;
  int inComment = false;
  ifstream inf(flname.c_str(), ios::in | ios::binary);
  if (!inf) { return ERRE; }
  s = "";
  while (!inf.eof()) {
    inf.get(c);
    if (inComment) {
      if (c == '\n') { inComment = false; }
    } else {
      if (c == '!') { inComment = true; }
      else if (isgraph(c)) { s += c; }
    }
  }
  inf.close();
  return NORM;
}

/*----- get base name of the filename ----------------------------------*/

string getBaseName(string &s) {
  int posDot, posSlsh, posBsls, posSep = -1;
  string ws, bnam;

  posSlsh = s.rfind('/');
  posBsls = s.rfind('\\');
  if (posSlsh != string::npos) {
    if (posBsls != string::npos) {
      if (posBsls > posSlsh) { posSep = posBsls; }
      else { posSep = posSlsh; }
    } else { posSep = posSlsh; }
  } else if (posBsls != string::npos) { posSep = posBsls; }
  ws = s.substr(posSep + 1);
  if ((posDot = ws.rfind('.')) != string::npos)
    { bnam = ws.substr(0, posDot); }
  else { bnam = ws; }
  return bnam;
}

/*----- get base part of the name end with strEnd ----------------------*/
string getBasePartEndWith(string nam, string strEnd) {
  int lnam, lstre;
  string ws;

  lnam = nam.length(); lstre = strEnd.length();
  if (lnam < lstre) { return getBaseName(nam); }
  if (nam.substr(lnam - lstre, lstre) != strEnd)
    { return getBaseName(nam); }
  ws = nam.substr(0, lnam - lstre) + '.';
  return getBaseName(ws);
}

/*----- G.C.M. and L.C.M. ----------------------------------------------*/

int getGCM(int x, int y) {
  int mx, im, gcm;
  if (x > y) { mx = x; } else { mx = y; }
  for (im = 2, gcm = 1; im <= mx; im++) {
    if (((x % im) == 0) && ((y % im) == 0)) { gcm = im; }
  }
  return gcm;
}

int getLCM(int x, int y) {
  int gcm, lcm;
  gcm = getGCM(x, y);
  lcm = x * y / gcm;
  return lcm;
}

/*----- file copy ------------------------------------------------------*/

int copyFile(string flnin, string flnout) {
  string s;
  ifstream ifs; ofstream ofs;
  ifs.open(flnin.c_str());
  if (!ifs) { cout << " Failed to read " << flnin << ".\n"; return ERRE; }
  ofs.open(flnout.c_str());
  if (!ofs) {
    cout << " Failed to write " << flnout << ".\n";
    ifs.close(); return ERRE;
  }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ofs.close();
  ifs.close();
  return NORM;
}

/*----- add 2 files into a file ----------------------------------------*/

int addToFile(string flnin1, string flnin2, string flnout) {
  string s;
  ifstream ifs1, ifs2; ofstream ofs;
  ofs.open(flnout.c_str());
  if (!ofs) { cout << " Failed to write " << flnout << ".\n"; return ERRE; }
  ifs1.open(flnin1.c_str());
  if (!ifs1) {
    cout << " Failed to read " << flnin1 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs1.eof()) {
    if (getOneLine(ifs1, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs1.close();
  ifs2.open(flnin2.c_str());
  if (!ifs2) {
    cout << " Failed to read " << flnin2 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs2.eof()) {
    if (getOneLine(ifs2, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs2.close();
  ofs.close();
  return NORM;
}

/*----- add 3 files into a file ----------------------------------------*/

int add3Files(string flnin1, string flnin2, string flnin3, string flnout) {
  string s;
  ifstream ifs1, ifs2, ifs3; ofstream ofs;
  ofs.open(flnout.c_str());
  if (!ofs) { cout << " Failed to write " << flnout << ".\n"; return ERRE; }

  ifs1.open(flnin1.c_str());   // ----- file 1 -----
  if (!ifs1) {
    cout << " Failed to read " << flnin1 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs1.eof()) {
    if (getOneLine(ifs1, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs1.close();

  ifs2.open(flnin2.c_str());   // ----- file 2 -----
  if (!ifs2) {
    cout << " Failed to read " << flnin2 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs2.eof()) {
    if (getOneLine(ifs2, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs2.close();

  ifs3.open(flnin3.c_str());   // ----- file 3 -----
  if (!ifs3) {
    cout << " Failed to read " << flnin3 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs3.eof()) {
    if (getOneLine(ifs3, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs3.close();

  ofs.close();
  return NORM;
}

/*------------------------------------------------------------------------
     List of temperatures
------------------------------------------------------------------------*/

tempList::tempList() { setDefault(); }            //---- (constructor) ---

void tempList::setDefault()                       //------- setDefault ---
  { setByRange(300., 1500., 100.); }

void tempList::setThermDefault() {              // --- setThermDefault ---
  clear();
  push_back(300.); push_back(400.); push_back(500.); push_back(600.);
  push_back(800.); push_back(1000.); push_back(1500.);
  sortElimDup();
}

void tempList::setPAC99Default() {              // --- setPAC99Default ---
  clear(); push_back(298.15);
  addRange(200., 1500., 100.); addRange(1600., 2800., 200.);
  addRange(3000., 3900., 300.); addRange(4200., 5000., 400.);
}

int tempList::readFrom(string flnam) {            //--------- readFrom ---
  string slin, key;
  vector<string> tkns;
  ifstream inf;
  inf.open(flnam.c_str());
  if (!inf) { return errms(FAILopen, flnam); }
  if (getLineApp(inf, slin) == NORM) {
    if (setBy(slin) == ERRE) { inf.close(); return ERRE; }
    while (getLineApp(inf, slin) == NORM) {
      if (addFrom(slin) == ERRE) { inf.close(); return ERRE; }
    }
  } else {
    return errms(EMPfile, flnam);
  }
  inf.close();
  return NORM;
}

int tempList::addFrom(string slin) {   // --- add from an input string ---
  string key;
  vector<string> tkns;
  splitToken(slin, " ,\t", tkns);
  if (tkns.empty()) { return errms(INVparN); }
  key = tkns[0];
  if (key == "tempRange") { return addRangeVS(tkns); }
  else if (key == "tempRecipRange") { return addRecipRangeVS(tkns); }
  else if (key == "tempGauChebGrd") { return addGauChebGrdVS(tkns); }
  else if (key == "tempList") { return addListVS(tkns); }
  return errms(INVkey, key);
}

int tempList::setBy(string slin)         // --- set by an input string ---
 { clear(); return addFrom(slin); }

                          // --- sort and eliminate duplicated element ---
void tempList::sortElimDup() {
  int id, jd, nelm = size();
  double d, dlast;
  vector<double> vd = *this;
  if (nelm < 2) { return; }
  for (id = 0; id < (nelm - 1); id++) {
    for (jd = (id + 1); jd < nelm; jd++) {
      if (vd[jd] < vd[id]) { d = vd[id]; vd[id] = vd[jd]; vd[jd] = d; }
    }
  }
  clear(); push_back(vd[0]); dlast = vd[0];
  for (id = 1; id < nelm; id++) {
    if ((vd[id] - dlast) > 1.e-10) { push_back(vd[id]); dlast = vd[id]; }
  }
}

                                    // --- add a range of temperatures ---
int tempList::addRange(double st, double et, double dt) {
  double t;
  if (dt < 1.e-6) { return errms(INVstep); }
  if ((st < 1.e-6) || (et < st)) { return errms(INVrange); }
  if (((et - st) / dt) > 1.e4) { return errms(INVrngMANY); }
  for (t = st; t < (et + dt * .1); t += dt) { push_back(t); }
  sortElimDup();
  return NORM;
}

                                       // --- add a range from strings ---
int tempList::addRangeVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempRange"
  double st, et, dt;
  if (vs.size() < 4) { return errms(INVparN); }
  st = atof(vs[1].c_str()); et = atof(vs[2].c_str());
  dt = atof(vs[3].c_str());
  return addRange(st, et, dt);
}

                                 // --- set by a range of temperatures ---
int tempList::setByRange(double st, double et, double dt)
 { clear(); return addRange(st, et, dt); }

                                    // --- set by a range from strings ---
int tempList::setByRangeVS(vector<string> &vs)
 { clear(); return addRangeVS(vs); }

                            // --- add a range from reciprocal numbers ---
int tempList::addRecipRange(double numr, double stt, double lst,
 double stp) {
  double rt;
  if (stp < 1.e-12) { return errms(INVstep); }
  if ((stt < 1.e-12) || (lst < stt)) { return errms(INVrange); }
  if (((lst - stt) / stp) > 1.e4) { return errms(INVrngMANY); }
  for (rt = stt; rt < (lst + stp * .1); rt += stp)
   { push_back(numr / rt); }
  sortElimDup();
  return NORM;
}

                 // --- add a range from reciprocal numbers in strings ---
int tempList::addRecipRangeVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempRecipRange"
  double numr, stt, lst, stp;
  if (vs.size() < 5) { return errms(INVparN); }
  numr = atof(vs[1].c_str()); stt = atof(vs[2].c_str());
  lst = atof(vs[3].c_str()); stp = atof(vs[4].c_str());
  return addRecipRange(numr, stt, lst, stp);
}

                         // --- set by a range from reciprocal numbers ---
int tempList::setByRecipRange(double numr, double stt, double lst,
 double stp)
 { clear(); return addRecipRange(numr, stt, lst, stp); }

              // --- set by a range from reciprocal numbers in strings ---
int tempList::setByRecipRangeVS(vector<string> &vs)
 { clear(); return addRecipRangeVS(vs); }

                                // --- add Gauss-Chebyshev grid points ---
int tempList::addGauChebGrd(double Tmin, double Tmax, int nT) {
  int i;
  double Ttil, Tinv, Tmini = 1. / Tmin, Tmaxi = 1. / Tmax;
  for (i = 0; i < nT; i++) {
    Ttil = cos((2. * i + 1.) / 2. / nT * M_PI);
    Tinv = (Ttil * (Tmaxi - Tmini) + (Tmaxi + Tmini)) / 2.;
    push_back(1. / Tinv);
  }
  sortElimDup();
  return NORM;
}

                     // --- add Gauss-Chebyshev grid points in strings ---
int tempList::addGauChebGrdVS(vector<string> &vs) {
  int nT;
  double Tmin, Tmax;
  if (vs.size() < 4) { return errms(INVparN); }
  Tmin = atof(vs[1].c_str()); Tmax = atof(vs[2].c_str());
  nT = atoi(vs[3].c_str());
  return addGauChebGrd(Tmin, Tmax, nT);
}

                             // --- set by Gauss-Chebyshev grid points ---
int tempList::setByGauChebGrd(double Tmin, double Tmax, int nT)
 { clear(); return addGauChebGrd(Tmin, Tmax, nT); }

                  // --- set by Gauss-Chebyshev grid points in strings ---
int tempList::setByGauChebGrdVS(vector<string> &vs)
 { clear(); return addGauChebGrdVS(vs); }

                                     // --- add a list of temperatures ---
int tempList::addList(vector<double> &vd) {
  int id, nd = vd.size();
  for (id = 0; id < nd; id++) {
    if (vd[id] < 1.e-6) { return errms(INVtemp); }
    push_back(vd[id]);
  }
  sortElimDup();
  return NORM;
}

                          // --- add a list of temperatures in strings ---
int tempList::addListVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempList"
  int is, ns = vs.size();
  vector<double> vd;
  vd.clear();
  for (is = 1; is < ns; is++) { vd.push_back(atof(vs[is].c_str())); }
  if (!vd.empty()) { return addList(vd); }
  return NORM;
}

                                  // --- set by a list of temperatures ---
int tempList::setByList(vector<double> &vd)
 { clear(); return addList(vd); }

                       // --- set by a list of temperatures in strings ---
int tempList::setByListVS(vector<string> &vs)
 { clear(); return addListVS(vs); }

ostream &tempList::print(ostream &os) {           //------------ print ---
  int itl, ntl = size();
  os << "  tempList";
  for (itl = 0; itl < ntl; itl++) { os << " " << (*this)[itl]; }
  os << endl;
  return os;
}

double tempList::maxTemp() {                      //---------- maxTemp ---
  int itl, ntl = size();
  double maxT = 0.;
  for (itl = 0; itl < ntl; itl++) {
    if ((*this)[itl] > maxT) { maxT = (*this)[itl]; }
  }
  return maxT;
}

                                                  // --- error message ---
int tempList::errms(errcode erc, string sv) {
  string msg = "tempList: ";
  switch(erc) {
    case FAILopen: msg += "Failed to open [" + sv + "]."; break;
    case EMPfile: msg += "Empty file [" + sv + "]."; break;
    case INVkey: msg += "Invalid key[" + sv + "]."; break;
    case INVstep: msg += "Invalid (too small) step input."; break;
    case INVrange: msg += "Invalid range input."; break;
    case INVrngMANY: msg += "Invalid range (too many) input."; break;
    case INVparN: msg += "Too small number of input parameters."; break;
    case INVtemp: msg += "Invalid temperature (< 0) input."; break;
  }
  cout << msg << endl;
  return ERRE;
}

ostream &operator<<(ostream &os, tempList &tl)    //------ operator << ---
  { return tl.print(os); }

