/*--------------------------------------------------------------------------
     "thermList.cc" - Read thermodata and output to .lst file
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: May   30, 2011
                                        edited: Sept. 11, 2013
                                   last edited: May   21, 2021
----------------------------------------------------------------------------
 usage: thermList input.ext [-new/-old]
--------------------------------------------------------------------------*/

#include "thermNASA.h"
using namespace std;

#define NTMP 7

double Tref = 298.15;
double Tlst[NTMP] = {300., 400., 500., 600., 800., 1000., 1500.};

/* ===== main ========================================================= */

int main(int argc, char *argv[]) {
  int i, numatm, flagOld = true, retc;
  double HRTref, SRref, CpRref;
  double HRT, SR, CpR;
  string flninp, flnbas, flnout, sopt, namspc, namelm;
  ofstream ofs;
  atmcCompSorted acs;
  atmcCompSorted::iterator a;
  thermTable thmCEA;
  thermTable::iterator p;

  showRevision();
  if ((argc < 2) || (argc > 3))
    { cout << "usage: thermList input.ext [-new/-old]\n"; return 1; }
  flninp = argv[1]; flnbas = getBaseName(flninp);
  flnout = flnbas + ".lst";
  if (argc == 3) {
    sopt = argv[2];
    if (sopt == "-new") { flagOld = false; }
    else if (sopt == "-old") { flagOld = true; }
    else { cout << "usage: thermList input.ext [-new/-old]\n"; return 1; }
  }
  if (flagOld) { retc = thmCEA.appendOldFrom(flninp); }
  else { retc = thmCEA.appendNewFrom(flninp); }
  if (retc == ERRE)
    { cout << "failed to read file [" << flninp << "].\n"; return 1; }

  ofs.open(flnout.c_str());
    if (!ofs) { cout << "failed to open [" << flnout << "].\n"; return 1; }
  ofs << "  UNITS:KJ\n";
  ofs << " Thermodata generated by thermNASA\n";
  ofs << " SPECIES       Hf       S    Cp 300     400     500     600     800     1000     1500     DATE        ELEMENTS\n";
  for (p = thmCEA.begin(); p != thmCEA.end(); p++) {
    namspc = p->second.nameSpc + "                  ";
    if (p->second.getThermFuncs(Tref, HRTref, SRref, CpRref) == ERRE)
      { cout << "calculation error.\n"; ofs.close(); return 1; }
    ofs << ' ' << namspc.substr(0, 10)
        << formatF(HRTref * RGAS * Tref / 1000., 8, 2) << "  "
        << formatF(SRref * RGAS, 7, 2) << "  ";
    for (i = 0; i < NTMP; i++) {
      if (p->second.getThermFuncs(Tlst[i], HRT, SR, CpR) == ERRE) {
        ofs << "   n.a. ";
      } else {
        ofs << formatF(CpR * RGAS, 7, 2) << ' ';
      }
    }
    ofs << namspc.substr(10, 8) << "tnasa    ";
    acs.setFrom(p->second.atmCmp);
    a = acs.begin();
    for (i = 0; i < 4; i++) {
      if (a != acs.end()) {
        namelm = elmt.canonSymbol(a->second.first);
        if (namelm.length() == 1) { namelm = namelm + " "; }
        else { namelm[1] = toupper(namelm[1]); }
        numatm = a->second.second;
        ofs << namelm << formatD(numatm, 3) << " ";
        a++;
      } else {
        ofs << "    0 ";
      }
    }
    ofs << "G 0\n";
  }
  ofs.close();
  return 0;
}

