/*--------------------------------------------------------------------------
     "thermNASA.h" - thermodynamic functions in NASA polynomial form
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: June  17, 2005
                                        edited: Sept. 27, 2013
                                   last edited: May   20, 2021
--------------------------------------------------------------------------*/

#ifndef _THERMNASA_H_
#define _THERMNASA_H_

#include "element.h"
using namespace std;

#define MXNASACOEF 8

// ***** thermNASA and related classes ***********************************

// --- coefNASAset: class of coefficients for single T interval ----------

class coefNASAset {
 public:
  int nCpCoef;   // 5 for old form and 7 (or 8) for new form
  double lowT, highT, dH298;
  double nx[MXNASACOEF], a[MXNASACOEF], b[2];
 public:                                     // --- member functions ---
  coefNASAset();
  void clear();
  int restoreOldFrom(vector<string> &vs, double deflT, double defhT);
  int restoreNewFrom(vector<string> &vs, double deflT, double defhT);
  int isTinInterval(double T);
  void calcThermFuncs(double T, double &HRT, double &SR, double &CpR);
  ostream &print(ostream &os);
 private:
  double getValOldCoef(const string &s);
  string fixOldCoefStr(const string &s);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, coefNASAset &cf);

// --- thermNASA: class of thermodynamic properties for single species ---

class thermNASA : public vector<coefNASAset> {
 public:                                         // --- data members ---
  int isUsed, phase;
  string nameSpc, comments, optID;
  atmcComp atmCmp;
  double wgtMol, H298;
 public:                                     // --- member functions ---
  thermNASA();
  thermNASA(const thermNASA &tm);
  void clear();
  int restoreOldFrom(vector<string> &vs, vector<double> &thT);
  int restoreNewFrom(vector<string> &vs, vector<double> &thT);
  int getThermFuncs(double T, double &HRT, double &SR, double &CpR);
  ostream &print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, thermNASA &th);

// --- thermTable: class of thermodynamic db table -----------------------

class thermTable : public map<string, thermNASA> {
 public:                                         // --- data members ---
  string comments;
  vector<double> thresT;
 public:                                     // --- member functions ---
  thermTable();
  void clear();
  void setM();           // set 'M' & 'm' in table
  int appendFrom(string flname);
  int appendOldFrom(string flname);
  int appendNewFrom(string flname);
  void setAllIsUsed();
  set<int> getUsedAtoms();
  ostream &print(ostream &os);
 private:
  int appendOld(string flname);
  int appendNew(string flname);
  int getThermLine(ifstream &ifs, string &s);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, thermTable &tt);

#endif

